/****************************************************************************************
|  Description: Buzzer driver source file
|    File Name: buzzer.c
|
|----------------------------------------------------------------------------------------
|                          C O P Y R I G H T
|----------------------------------------------------------------------------------------
|     Copyright (c) 2012 by Feaser     http://www.feaser.com     All rights reserved
|
|   This software is the property of Feaser. The source and compiled code may only be 
|        used on the Olimex EM32G880F128-STK board (http://www.olimex.com/dev).
| 
|      This copyright notice may not be removed from the source code nor changed.
|
|   This software has been carefully tested, but is not guaranteed for any particular
| purpose. The author does not offer any warranties and does not guarantee the accuracy,
|   adequacy, or completeness of the software and is not responsible for any errors or
|              omissions or the results obtained from use of the software.
|
****************************************************************************************/

/****************************************************************************************
* Include files
****************************************************************************************/
#include "header.h"                                   /* generic header file           */
#include "efm32_gpio.h"                               /* EFM32 GPIO module             */


/****************************************************************************************
* Macro definitions
****************************************************************************************/
/* specify the period time at which function BuzzerUpdate() is called by the program */
#define BUZZER_UPDATE_PERIOD_MS     (1)


/****************************************************************************************
* Local data declarations
****************************************************************************************/
static unsigned char buzzerOn;
static unsigned char buzzerTimeCnt;


/****************************************************************************************
** NAME:           BuzzerInit
** PARAMETER:      none
** RETURN VALUE:   none
** DESCRIPTION:    Initializes the port pins connected to the buzzer.
**
****************************************************************************************/
void BuzzerInit(void)
{
  /* configure PA12 and PA13 as push-pull digital outputs */
  GPIO_PinModeSet(gpioPortA, 12, gpioModePushPullDrive, 0);
  GPIO_PinModeSet(gpioPortA, 13, gpioModePushPullDrive, 0);
  /* set port A outputs to drive up to 20 mA */
  GPIO_DriveModeSet(gpioPortA, gpioDriveModeHigh);
  /* buzzer if off by default */
  buzzerOn = FALSE;
  buzzerTimeCnt = 0;
} /*** end of BuzzerInit ***/


/****************************************************************************************
** NAME:           BuzzerOn
** PARAMETER:      duration time in milliseconds that the buzzer should be enabled.
** RETURN VALUE:   none
** DESCRIPTION:    Enables the buzzer for the time specified.
**
****************************************************************************************/
void BuzzerOn(unsigned short duration)
{
  /* activate the buzzer */
  buzzerTimeCnt = duration / BUZZER_UPDATE_PERIOD_MS;
  buzzerOn = TRUE;
} /*** end of BuzzerOn ***/


/****************************************************************************************
** NAME:           BuzzerOff
** PARAMETER:      none
** RETURN VALUE:   none
** DESCRIPTION:    Turns the buzzer off.
**
****************************************************************************************/
void BuzzerOff(void)
{
  /* deactivate the buzzer */
  buzzerOn = FALSE;
  buzzerTimeCnt = 0;
} /*** end of BuzzerOff ***/


/****************************************************************************************
** NAME:           BuzzerUpdate
** PARAMETER:      none
** RETURN VALUE:   none
** DESCRIPTION:    Update function that controls the buzzer. Should be called at a
**                 1ms period by the program.
**
****************************************************************************************/
void BuzzerUpdate(void)
{
  /* handle the timer counter */
  if (buzzerTimeCnt == 0)
  {
    buzzerOn = FALSE;
  }
  else
  {
    buzzerTimeCnt--;
  }

  /* handle the buzzer */
  if (buzzerOn == TRUE)
  {
    /* buzzer on so generate blockwave with fixed frequency to make some noise */
    if (GPIO_PinOutGet(gpioPortA, 12) == 0)
    {
      GPIO_PinOutSet(gpioPortA, 12);
      GPIO_PinOutClear(gpioPortA, 13);
    }
    else
    {
      GPIO_PinOutClear(gpioPortA, 12);
      GPIO_PinOutSet(gpioPortA, 13);
    }
  }
  else
  {
    GPIO_PinOutClear(gpioPortA, 12);
    GPIO_PinOutClear(gpioPortA, 13);
  }
} /*** end of BuzzerUpdate ***/


/************************************ end of buzzer.c **********************************/


