/****************************************************************************************
|  Description: Operating system source file
|    File Name: os.c
|
|----------------------------------------------------------------------------------------
|                          C O P Y R I G H T
|----------------------------------------------------------------------------------------
|     Copyright (c) 2012 by Feaser     http://www.feaser.com     All rights reserved
|
|   This software is the property of Feaser. The source and compiled code may only be 
|        used on the Olimex EM32G880F128-STK board (http://www.olimex.com/dev).
| 
|      This copyright notice may not be removed from the source code nor changed.
|
|   This software has been carefully tested, but is not guaranteed for any particular
| purpose. The author does not offer any warranties and does not guarantee the accuracy,
|   adequacy, or completeness of the software and is not responsible for any errors or
|              omissions or the results obtained from use of the software.
|
|***************************************************************************************/


/****************************************************************************************
* Feaser non-preemptive multitasking operating system
* 
* Introduction
* This operating system is an easy to use and easy to integrate multitasking operating
* system. The scheduler is non-preemptive and priority based. 
*
* Its memory footprint is extremely small, making it suitable for microcontrollers
* starting with 2 kbytes of ROM and 128 bytes of RAM. It enables small microcontroller 
* systems to use multitasking. Compared to the primitive foreground/background method, 
* also called superloop, this operating system results in improved real-time performance.
*
* The current version supports up to 16 tasks. When more tasks are needed and the
* microcontroller has sufficient RAM resources (2 kbytes or more), consider using a
* preemptive multitasking operating system such as Micrium's MicroC/OS-II available from
* http://www.micrium.com or FreeRTOS available from http://www.freertos.org
*
* Usage example
*     void main(void)
*     {
*       OsInit();
*       OsTaskCreate(AppStartTask, 0, 0);
*       EnableInterrupts();
*       OsStart();
*     }
*
*     void AppStartTask(void)
*     {
*       OsTaskCreate(AppTask1, 2, OS_TICKS_PER_SEC/2);
*       OsTaskCreate(AppTask2, 4, 0);
*       OsTaskDelete(OS_PRIO_SELF);
*     }
*
*     void AppTask1(void)
*     {
*       LedToggle(1);
*     }
*
*     void AppTask2(void)
*     {
*       LedToggle(2);
*       OsTaskDelay(OS_PRIO_SELF, OS_TICKS_PER_SEC/4);
*     }
*
* Scheduler internals
* - The scheduler is non-preemptive and priority based.
* - Each task has its unique priority number between 0 and OS_MAX_TASKS-1.
* - A higher priority number represents a higher priority.
* - Four bits are needed to store the priority number:
*      ---------------
*     |*|*|*|*|y|y|x|x|
*      ---------------
* - When a task is ready-to-run, its associated bit is set to 1 in osTaskReadyTable[]:
*     |**|**|**|**|03|02|01|00| (group 0)
*     |**|**|**|**|07|06|05|04| (group 1)
*     |**|**|**|**|11|10|09|08| (group 2)
*     |**|**|**|**|15|14|13|12| (group 3)
* - To efficiently determine which flags belongs to what prioritynumber and vise versa,
*   all task ready flags are spread over 4 groups, with each group holding 4 bits.
* - A group flag in osTaskReadyGroup is set when a group has a ready-to-run task:
*      -----------------------------------------------------------
*     |unused|unused|unused|unused|group 3|group 2|group 1|group 0| 
*      -----------------------------------------------------------
* - Bits 2 and 3 in the priority number determine the group (y) and bits 0 and 1 
*   determine the priority within the group (x).
* - Relation between priority number and x, y is:
*     priority = (y * 4) + x;
****************************************************************************************/


/****************************************************************************************
* Include files
****************************************************************************************/
#include "os.h"                                      /* operating system header file   */


/****************************************************************************************
* Defines
****************************************************************************************/
/* Internally used as an indicator and validator that a task priority value is invalid.
 */
#define OS_PRIO_INVALID      (255)


/****************************************************************************************
* Type definitions
****************************************************************************************/
/* Each task is assigned its own task control data structure to keep track of its state.
 */
typedef struct
{
  /* For dormant tasks, the function pointer is set to a NULL pointer. This
   * NULL pointer is additionally used to determine if a task has already
   * been created.
   */
  void           (*function)(void);
  /* Upon tasks creation, a period time in units of time ticks is specified.
   * If this value is non-zero, the task make the tasks ready-to-run each time 
   * the specified number of time ticks elapsed.
   */
  unsigned short period;
  /* This value is decremented with each time tick. The moment it reaches a 
   * zero value, the task is made ready to run. For periodic tasks, the value
   * is then reloaded with the task's period. Explicitely setting this
   * structure element to zero, suspends a tasks.
   */
  unsigned short expires;
} tOsTaskCtrl;


/****************************************************************************************
* Function prototypes
****************************************************************************************/
static void          OsTaskSetReady(unsigned char prio);
static void          OsTaskResetReady(unsigned char prio);
static unsigned char OsTaskGetReady(void);


/****************************************************************************************
* Local data declarations
****************************************************************************************/
/* All possibly supported task controls are grouped in this array. The handle
 * of a task control is the index into this array.
 */
static tOsTaskCtrl   osTaskCtrl[OS_MAX_TASKS];

/* A group flag is set when a task group has a ready-to-run task */
static unsigned char osTaskReadyGroup;

/* When a task is running, this variable holds the priority value of the
 * running task, otherwise OS_PRIO_INVALID. This variable is necessary to 
 * implement the OS_PRIO_SELF feature. 
 */

static unsigned char osCurrTaskPrio;

/* When a task is ready-to-run, its associated bit is set to 1 in this array */
static unsigned char osTaskReadyTable[(OS_MAX_TASKS/4)];

/****************************************************************************************
* Local constant declarations
****************************************************************************************/
/* Lookup table to determine x and y of the highest priorty task ready-to-run */
static const unsigned char osTaskReadyUnmap[] =
{
  0, /* %0000 -> invalid */
  0, /* %0001 -> 0       */
  1, /* %0010 -> 1       */
  1, /* %0011 -> 1       */
  2, /* %0100 -> 2       */
  2, /* %0101 -> 2       */
  2, /* %0110 -> 2       */
  2, /* %0111 -> 2       */
  3, /* %1000 -> 3       */
  3, /* %1001 -> 3       */
  3, /* %1010 -> 3       */
  3, /* %1011 -> 3       */
  3, /* %1100 -> 3       */
  3, /* %1101 -> 3       */
  3, /* %1110 -> 3       */
  3  /* %1111 -> 3       */
};


/****************************************************************************************
** NAME:           OsInit
** PARAMETER:      none
** RETURN VALUE:   none
** DESCRIPTION:    Initializes the operating system. Function should be called once
**                 before any of the other operating system management functions are
**                 called. Function is non-reentrant so do not call from an ISR.
**
****************************************************************************************/
void OsInit(void)
{
  unsigned char i;

  /* init internal data structures */
  osTaskReadyGroup = 0;
  osCurrTaskPrio = OS_PRIO_INVALID;
  for (i = 0; i < (sizeof(osTaskReadyTable)/sizeof(osTaskReadyTable[0])); i++)
  {
    osTaskReadyTable[i] = 0;
  }

  /* init all task controls */
  for (i = 0; i < (OS_MAX_TASKS); i++)
  {
    osTaskCtrl[i].function = (void *)0;
    osTaskCtrl[i].period = 0;
    osTaskCtrl[i].expires = 0;
  }

  /* init the time tick */
  HwOsTimeTickInit();
} /*** end of OsInit ***/


/****************************************************************************************
** NAME:           OsTimeTick
** PARAMETER:      none
** RETURN VALUE:   none
** DESCRIPTION:    The time tick drives the operating system. Function should be called 
**                 periodically at a 1/OS_TICKS_PER_SEC frequency. Typically called by a 
**                 timer interrupt.
**
****************************************************************************************/
void OsTimeTick(void)
{
  unsigned char i;

  /* loop through all tasks */
  for (i = 0; i < (OS_MAX_TASKS); i++)
  {
    OS_ENTER_CRITICAL();
    /* update tasks expiration time */
    if (osTaskCtrl[i].expires > 0)
    {
      /* update expiration time */
      osTaskCtrl[i].expires--;
      /* set ready and reload when expired */
      if (osTaskCtrl[i].expires == 0)
      {
        OsTaskSetReady(i);  
        osTaskCtrl[i].expires = osTaskCtrl[i].period;
      }
    }
    OS_EXIT_CRITICAL();
  }
} /*** end of OsTimeTick ***/


/****************************************************************************************
** NAME:           OsStart
** PARAMETER:      none
** RETURN VALUE:   none
** DESCRIPTION:    Starts the operating system. Function should be called after OsInit() 
**                 and at least one task was created. Note that this function does not
**                 return.
**
****************************************************************************************/
void OsStart(void)
{
  unsigned char prio;
  void (*task_function)(void);

  /* infinite scheduling loop */
  for (;;)
  {
    /* get the highest priority task that is ready-to-run */
    prio = OsTaskGetReady();

    /* at least one task ready-to-run? */
    if (prio != OS_PRIO_INVALID)
    {
      OS_ENTER_CRITICAL();
      task_function  = osTaskCtrl[prio].function;
      osCurrTaskPrio = prio;
      OS_EXIT_CRITICAL();

      /* reset ready-flag. should be called before the task function call, otherwise
       * a task cannot ready itself.
       */
      OsTaskResetReady(prio);

      /* call the task if valid */
      if (task_function != (void *)0)
      {
        task_function();
      }

      OS_ENTER_CRITICAL();
      osCurrTaskPrio = OS_PRIO_INVALID;
      OS_EXIT_CRITICAL();
    }
  }
} /*** end of OsStart ***/


/****************************************************************************************
** NAME:           OsTaskCreate
** PARAMETER:      task   Pointer to the task's function
**                 prio   Priority of the task.
**                 period Task cycle time in time ticks.
** RETURN VALUE:   OS_TRUE if the task was created, OS_FALSE otherwise.
** DESCRIPTION:    Places a task under the operating system's task control. The period
**                 parameter specifies if the task should be called periodically, 
**                 measured in time ticks. Use 0 to create a non-periodic task. 
**                 Priorities 0 to OS_MAX_TASKS-1 are  available. A higher priority 
**                 number relates to a higher task priority.
**
****************************************************************************************/
unsigned char OsTaskCreate(void (*task)(void), unsigned char prio, unsigned short period)
{
  void (*task_function)(void);

  OS_ENTER_CRITICAL();
  /* read out task function pointer */
  task_function = osTaskCtrl[prio].function;
  OS_EXIT_CRITICAL();

  /* validate parameters */
  if ( (prio >= OS_MAX_TASKS) || (task_function != (void *)0) )
  {
    return OS_FALSE;
  }

  OS_ENTER_CRITICAL();
  /* create task by setting its parameters */
  osTaskCtrl[prio].function = task;
  osTaskCtrl[prio].period = period;
  osTaskCtrl[prio].expires = period;
  OS_EXIT_CRITICAL();

  /* each task is ready-to-run after its creation */
  OsTaskSetReady(prio);

  return OS_TRUE;
} /*** end of OsTaskCreate ***/


/****************************************************************************************
** NAME:           OsTaskDelete
** PARAMETER:      prio   Priority of the task that should be deleted.
** RETURN VALUE:   OS_TRUE if the task was deleted, OS_FALSE otherwise.
** DESCRIPTION:    Removes a task from the operating system's task control. A task can 
**                 delete itself by specifying OS_PRIO_SELF as the parameter.
**
****************************************************************************************/
unsigned char OsTaskDelete(unsigned char prio)
{
  void (*task_function)(void);

  /* set priority in case a task wants to delete itself */
  if (prio == OS_PRIO_SELF)
  {
    OS_ENTER_CRITICAL();
    prio = osCurrTaskPrio;
    OS_EXIT_CRITICAL();
  }

  OS_ENTER_CRITICAL();
  /* read out task function pointer */
  task_function = osTaskCtrl[prio].function;
  OS_EXIT_CRITICAL();

  /* validate parameters */
  if ( (prio >= OS_MAX_TASKS) || (task_function == (void *)0))
  {
    return OS_FALSE;
  }

  OS_ENTER_CRITICAL();
  /* set to dormant state and set function pointer to a null-pointer */
  osTaskCtrl[prio].function = (void *)0;
  osTaskCtrl[prio].period = 0;
  osTaskCtrl[prio].expires = 0;
  OS_EXIT_CRITICAL();

  /* reset ready-flag */
  OsTaskResetReady(prio);

  return OS_TRUE;
} /*** end of OsTaskDelete ***/


/****************************************************************************************
** NAME:           OsTaskDelay
** PARAMETER:      prio   Priority of the task that should be delayed.
**                 delay  Delay time in time ticks.
** RETURN VALUE:   OS_TRUE if the task was delayed, OS_FALSE otherwise.
** DESCRIPTION:    Delays a task for the specified amount of time ticks. A task can delay
**                 itself by specifying OS_PRIO_SELF. Calling this function with the 
**                 delay parameter set to 0, completely suspends the task. A suspended 
**                 task can only be resumed by calling function OSTaskReady.
**
****************************************************************************************/
unsigned char OsTaskDelay(unsigned char prio, unsigned short delay)
{
  void (*task_function)(void);

  /* set priority in case a task wants to delete itself */
  if (prio == OS_PRIO_SELF)
  {
    OS_ENTER_CRITICAL();
    prio = osCurrTaskPrio;
    OS_EXIT_CRITICAL();
  }

  OS_ENTER_CRITICAL();
  /* read out task function pointer */
  task_function = osTaskCtrl[prio].function;
  OS_EXIT_CRITICAL();

  /* validate parameters */
  if ( (prio >= OS_MAX_TASKS) || (task_function == (void *)0))
  {
    return OS_FALSE;
  }

  OS_ENTER_CRITICAL();
  /* update the expiration time of the task */
  osTaskCtrl[prio].expires = delay;
  OS_EXIT_CRITICAL();

  return OS_TRUE;
} /*** end of OsTaskDelay ***/


/****************************************************************************************
** NAME:           OsTaskReady
** PARAMETER:      prio Priority of the task that should be made ready-to-run.
** RETURN VALUE:   OS_TRUE is the task was successfully readied, OS_FALSE otherwise.
** DESCRIPTION:    Forces a task into the ready-to-run state. A task can keep itself 
**                 ready by specifying OS_PRIO_SELF. If a periodic task is forced into 
**                 the ready-to-run state, its period also restarts.
**
****************************************************************************************/
unsigned char OsTaskReady(unsigned char prio)
{
  void (*task_function)(void);

  /* set priority in case a task wants to delete itself */
  if (prio == OS_PRIO_SELF)
  {
    OS_ENTER_CRITICAL();
    prio = osCurrTaskPrio;
    OS_EXIT_CRITICAL();
  }

  OS_ENTER_CRITICAL();
  /* read out task function pointer */
  task_function = osTaskCtrl[prio].function;
  OS_EXIT_CRITICAL();

  /* validate parameters */
  if ( (prio >= OS_MAX_TASKS) || (task_function == (void *)0))
  {
    return OS_FALSE;
  }

  OS_ENTER_CRITICAL();
  /* reload periodic tasks */
  if (osTaskCtrl[prio].period != 0)
  {
    osTaskCtrl[prio].expires = osTaskCtrl[prio].period;
  }
  OS_EXIT_CRITICAL();

  /* make the task ready-to-run */
  OsTaskSetReady(prio);

  return OS_TRUE;
} /*** end of OsTaskDelay ***/


/****************************************************************************************
** NAME:           OsTaskSetReady
** PARAMETER:      prio Selects the task is should be made ready-to-run.
** RETURN VALUE:   none
** DESCRIPTION:    Places a task in the ready-to-run state.
**
****************************************************************************************/
static void OsTaskSetReady(unsigned char prio)
{
  unsigned char x, y;  
  
  /* extract x and y from priority number */
  y = (prio >> 2u) & 0x03;
  x = prio & 0x03;
  OS_ENTER_CRITICAL();
  /* set bit in table using the group as the table index */ 
  osTaskReadyTable[y] |= (1 << x);
  /* set group bit */
  osTaskReadyGroup |= (1 << y);
  OS_EXIT_CRITICAL();
} /*** end of OsTaskSetReady ***/


/****************************************************************************************
** NAME:           OsTaskResetReady
** PARAMETER:      prio Selects the task is should be placed into the waiting state.
** RETURN VALUE:   none
** DESCRIPTION:    Places a task in the waiting state.
**
****************************************************************************************/
static void OsTaskResetReady(unsigned char prio)
{
  unsigned char x, y;  
  
  /* extract x and y from priority number */
  y = (prio >> 2) & 0x03;
  x = prio & 0x03;
  OS_ENTER_CRITICAL();
  /* reset bit in table using the group as the table index */ 
  osTaskReadyTable[y] &= ~(1u << x);
  /* if no other tasks in the group are ready, reset the group bit */
  if (osTaskReadyTable[y] == 0)
  {
    osTaskReadyGroup &= ~(1u << y);
  }
  OS_EXIT_CRITICAL();
} /*** end of OsTaskResetReady ***/ 


/****************************************************************************************
** NAME:           OsTaskGetReady
** PARAMETER:      none
** RETURN VALUE:   The priority of the highest ready-to-run task or OS_PRIO_INVALID if 
**                 all waiting.
** DESCRIPTION:    Obtains the priority number of the task with the highest priority 
**                 that is in the ready-to-run state.
**
****************************************************************************************/
static unsigned char OsTaskGetReady(void)
{
  unsigned char x,y;

  OS_ENTER_CRITICAL();
  /* detect if any task is ready first */
  if (osTaskReadyGroup == 0)
  {
    OS_EXIT_CRITICAL();
    return OS_PRIO_INVALID;
  }
  /* convert group bit mask to group number */
  y = osTaskReadyUnmap[osTaskReadyGroup];
  /* get associated group task number */
  x = osTaskReadyUnmap[osTaskReadyTable[y]];
  OS_EXIT_CRITICAL();
  /* construct priority number from x and y and return it */
  return (unsigned char)((y << 2) | x);
} /*** end of OsTaskGetReady ***/


/*********************************** end of os.c ***************************************/
